from pyparagraph import TextWrap

args = {
    "font": "Lato",
    "font-size": "24px",
    "width": "600"
}

text = """Now we are engaged in a great civil war, testing whether that nation, or any nation so conceived and so dedicated, can long endure. We are met on a great battle-field of that war. We have come to dedicate a portion of that field, as a final resting place for those who here gave their lives that that nation might live. It is altogether fitting and proper that we should do this."""

# Render to a file of the given name
textwrap = TextWrap.text_wrapper(args)
textwrap.render_image(text, filename="gettysburg.png", margin=20)


# Render to a PIL Image:

from PIL import Image
mine = Image.new("RGBA", (800,800), "white")

textwrap.render_image(text, img=mine)
mine.save("gettysburg2.png")



from collections import deque
class Node:
    def __init__(self, data):
        assert data is not None
        self.data = data
    def __repr__(self):
        return "<Node: %s>"%repr(self.data)
    def __str__(self):
        return "<Node: %s>"%repr(self.data)
    def __lt__(self, obj):
        return self.data.demerits < obj.data.demerits

class LinkedList(deque):
    def appendleft(self, node):
        node.prev = None
        if self:
            node.next = self[0]
            self[0].prev = node
        else:
            node.next = None
        deque.appendleft(self, node)

    def insert(self, i, new_node):
        if i==0:
            self.appendleft(new_node)
        else:
            old  = self[i-1]
            if i==len(self):
                new_node.next = None
            else:
                new_node.next = self[i]
                self[i].prev = new_node
            new_node.prev = old
            old.next = new_node
            deque.insert(self, i, new_node)

    def append(self, new_node):
        i = len(self)
        self.insert(i, new_node)

    def insertBefore(self, existing, new_node):
        i = self.index(existing)
        self.insert(i, new_node)

    def remove(self, node):
        i = self.index(node)
        if node.prev:
            node.prev.next = node.next
        if node.next:
            node.next.prev = node.prev
        deque.remove(self, node)




#!/usr/bin/env python3
import re

try:
    from hyphen import Hyphenator
    have_hyphenate = True
except:
    have_hyphenate = False

def pairwise(iterable):
    """list(pairwise([1,2,3,4,5,6])) --> [(1,2), (3,4), (5,6)]"""
    a = iter(iterable)
    return zip(a, a)

class HBox:
    def __init__(self, value, font):
        self.value = value
        self.font = font
        self.width = self.font.getlength(value)
    def __repr__(self):
        return f'HBox(value={repr(self.value)})'

class Penalty:
    def __init__(self, width, penalty, flagged):
        self.width = width
        self.penalty = penalty
        self.flagged = flagged
    def __repr__(self):
        return f'Penalty(width={repr(self.width)}, penalty={repr(self.penalty)}, flagged={repr(self.flagged)})'

class Glue:
    def __init__(self, width, stretch, shrink):
        self.width = width
        self.stretch = stretch
        self.shrink = shrink
    def __repr__(self):
        return f'Glue(width={repr(self.width)}, stretch={repr(self.stretch)}, shrink={repr(self.shrink)})'
    def __add__(self, other):
        return Glue(self.width+other.width, self.stretch+other.stretch, self.shrink+other.shrink)

Infinity = 10**10   # 10000 fails on some not uncommonly cramped sizes

def fix_hyphenation(words):
    rv = []
    for word in words:
        if not word[0].isalpha() and rv:
            rv[-1]+= word
        else:
            rv.append(word)
    return rv

# Handle a few characters that we want to be able to break after,
# even mid-word.  We don't break cases where the latter part is a number. 
# So "1/2" and "1-2" will not be split, but "3-part" and "Frost/Nixon" 
# may be split after the punctuation symbol.
#
# And there's a look-before to prevent breaking if the previous character
# is a <, so we don't break "</" "span>" mid-tag and similar.

wordbreak_chars = "-—–/>"
wordbreak_char_re = re.compile("(?<!<)([%s]+)(?!$|[0-9])"%wordbreak_chars)

#
# Breaking spaces: the whitespace spaces that should break words.
#

breaking_spaces = " \u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200A\u200B"
breaking_re = re.compile(f"([{breaking_spaces}])")

class BaseWordSplitter:
    def __init__(self, args, font=None):
        self.font = font
        self.hyphen_penalty = args.hyphen_penalty
        self.midbreak_penalty = args.midbreak_penalty
        self.hyphenate = args.hyphenate
        self.nobreak = 0  # Do not create breaks if > 0
        if self.hyphenate:
            self.hyphenator = Hyphenator('en_US')

    def split(self, text):
        # Canonicalize DOS and old Macintosh line endings to \n
        text = text.replace("\r\n", "\n")
        text = text.replace("\r", "\n")
        # Strip a single trailing \n if needed
        if text[-1] == "\n":
            text = text[:-1]
        paragraphs = text.split("\n")
        rv = []
        for paragraph in paragraphs:
            rv.append(self.split_paragraph(paragraph))
        return rv[0]

    def add_space(self, space=" "):
        """Depending on nobreak, adds a breaking or non-breaking space"""
        if self.nobreak:
            self.nodes.append(Penalty(0, Infinity, 0))
        self.nodes.append(Glue(0, self.font.get_space_width(space), 0))
        self.nodes.append(Penalty(0, Infinity if self.nobreak else 0, 0))
        self.nodes.append(Glue(self.font.get_space_width(space), self.font.get_space_stretch(space), self.font.get_space_shrink(space)))

    def add_newline(self):
        """Adds the newline nodes"""
        self.nodes.append(Glue(0, Infinity, 0))
        self.nodes.append(Penalty(0, -Infinity, 1))

    def add_nodes_from_part(self, part):
        """Handle a few characters that we want to be able to break after,
        even mid-word. self.midbreak_penalty is used to minimize these, but
        mid-morning could be split at mid-, for instance. See wordbreak_char_re
        comment above."""

        parts = wordbreak_char_re.split(part)
        *head, tail = parts
        for p, c in pairwise(head):
            self.nodes.append(HBox(p+c, self.active_font))
            self.nodes.append(Penalty(0, Infinity if self.nobreak else self.midbreak_penalty, 1))
        self.nodes.append(HBox(tail, self.active_font))

    def add_nodes_from_word(self, word):
        """Split out the nodes from a (possibly hyphenated) word, including
        splitting it further than the original whitespace splitting if
        needed."""

        if len(word)==1 and breaking_re.match(word):
            self.add_space(word)
            return

        if self.hyphenate and len(word)>6:
            words = self.hyphenator.syllables(word)
            words = fix_hyphenation(words)
        else:
            words = [word]

        *head, tail = words
        
        for part in head:
            self.add_nodes_from_part(part)
            self.nodes.append(Penalty(self.font.get_hyphen_width(), Infinity if self.nobreak else self.hyphen_penalty, 1))

        self.add_nodes_from_part(tail)


    def add_words_from_string(self, string):
        if not string:
            return
        if string:
            need_start = string[0]==" "
            need_end = string[-1]==" " and len(string)>1
            if need_start:
                string = string[1:]
            if need_end:
                string = string[:-1]
        all_words = breaking_re.split(string)

        if need_start:
            self.add_space()

        for word in all_words:
            self.add_nodes_from_word(word)

        if need_end:
            self.add_space()


from html.parser import HTMLParser
class RichWordSplitter(BaseWordSplitter, HTMLParser):
    def __init__(self, *args, **kwargs):
        BaseWordSplitter.__init__(self, *args, **kwargs)
        self.bold = 0
        self.italic = 0
        self.set_active_font()
        self.nodes = []
        HTMLParser.__init__(self)

    def split_paragraph(self, text):
        self.nodes = []
        self.feed(text)
        self.add_newline()
        return self.nodes

    def set_active_font(self):
        if self.bold and self.italic:
            self.active_font = self.font.bolditalicfont
        elif self.bold:
            self.active_font = self.font.boldfont
        elif self.italic:
            self.active_font = self.font.italicfont
        else:
            self.active_font = self.font.font
    def handle_starttag(self, tag, attrs):
        if tag.lower() == 'b':
            self.bold += 1
            self.set_active_font()
        elif tag.lower() == 'i':
            self.italic += 1
            self.set_active_font()
        elif tag.lower() == 'nb':
            self.nobreak += 1
        else:
            self.nobreak += 1
            self.handle_data(self.get_starttag_text())
            self.nobreak -= 1
    def push_nodes(self, string):
        self.add_words_from_string(string)

    def handle_endtag(self, tag):
        if tag.lower() == 'b':
            self.bold -= 1
            self.set_active_font()
        elif tag.lower() == 'i':
            self.italic -= 1
            self.set_active_font()
        elif tag.lower() == 'nb':
            self.nobreak -= 1
        else:
            self.nobreak += 1
            self.handle_data(f"</{tag}>")
            self.nobreak -= 1
    def handle_data(self, data):
        data = data.split("\n")
        for i, datum in enumerate(data):
            self.push_nodes(datum)
            if i != len(data) - 1:
                self.add_newline()

class BasicWordSplitter(BaseWordSplitter):
    def split_paragraph(self, text):
        self.nodes = []
        self.active_font = self.font.font

        self.add_words_from_string(text)
        self.add_newline()

        return self.nodes



#!/usr/bin/env python3
import os
import logging
import sys

from PIL import ImageFont

try:
    from matplotlib import font_manager
    logging.getLogger("matplotlib").setLevel(logging.INFO)
except:
    font_manager = None

class NonProportional:
    """A dummy class used for text-mode/console rendering"""
    @classmethod
    def getlength(cls, string):
        return len(string)

class Font:
    """Tries to lookup the specified font, either as a TTF/OTF file
    or an OS-installed font family. Looks for bold, italic, and 
    bold-italic variants, and loads them all; handles finding metrics
    for strings/characters."""
    def __init__(self, filename, size="12pt", linespacing=1.4, boldfilename=None, italicfilename=None, bolditalicfilename=None):
        if size.endswith("pt"):
            size = round(4 * float(size[:-2])/3)
        elif size.endswith("px"):
            size = round(float(size[:-2]))
        else:
            raise Exception("Size must specify px or pt, e.g. 12pt or 16px")

        self.size = size
        self.space_cache = {}
        self.linespacing = linespacing

        self.filename = filename

        if filename is not NonProportional:
            def find_replace(fname, look_for_list):
                """Seek out Bold, Italic, and, BoldItalic equivalents if the primary is -Regular or -Medium.
                If the font is Times-Regular.ttf, look for Times-Bold.ttf, Times-Italic.ttf, and such
                If times-regular.ttf, look for times-bold.ttf etc. Return a match, or None if not found.

                This tries all sorts of permutations: '-Bold-Italic', 'Bold-Italic', ' Bold Italic', etc."""

                dirname, filename = os.path.split(fname)
                for template in "Regular", "Medium":
                    for part in ("-"+template, " "+template, template):
                        for look_template in look_for_list:
                            for look_for in (look_template, look_template.replace("-", " "), look_template.lstrip("-")):
                                if part in filename:
                                    test = filename.replace(part, look_for)
                                elif part.lower() in filename:
                                    test = filename.replace(part.lower(), look_for.lower())
                                else:
                                    continue
                                test = os.path.join(dirname, test)
                                if os.path.isfile(test):
                                    return test

                # Now try Times.ttf -> TimesBold.ttf, Times Bold.ttf, etc.
                if "." in filename:
                    base, ext = filename.rsplit(".", 1)
                    for look_for in look_for_list:
                        for look_for_case in (look_for, look_for.lower(), look_for[1:], look_for[1:].lower(), " "+look_for[1:], " "+look_for[1:].lower()):
                            test = os.path.join(dirname, "%s%s.%s"%(base, look_for_case, ext))
                            if os.path.isfile(test):
                                return test
                        
                return None

            # Filename t
            replacements = {
                "boldfilename": ("-Bold",),
                "italicfilename": ("-Italic",),
                "bolditalicfilename": ("-BoldItalic", "-Bold-Italic", "-Bold Italic"),
            }

            print("Loading font")
            if filename is None and font_manager is None:
                logging.warning("Cannot load default font; falling back to PIL default. 'pip install matplotlib' to upgrade font handling.")
                self.font = ImageFont.load_default()
            else:
                if filename is None:
                    for name in ('Liberation Sans', 'Calibri', 'Lucida Grande', 'Helvetica', 'Verdana', 'Arial', 'Noto Sans', 'Ubuntu', 'DejaVu Sans', 'sans-serif'):
                        try:
                            font = font_manager.FontProperties(family=name)
                            self.filename = font_manager.findfont(font, fallback_to_default=False)
                            break
                        except:
                            logging.debug("No font %s", name)
                    else:
                        logging.error("No font found, not even sans-serif. Use --font NAME with an installed font or TTF/OTF font filename.")
                    logging.info("Using font: %s", self.filename)

                if not os.path.isfile(self.filename):
                    if font_manager is None:
                        logging.warning("Font is not a filename and no font manager available. 'pip install matplotlib' to find fonts by family.")
                        logging.error("Not a file: %s", self.filename)
                        sys.exit(-1)
                    else:
                        try:
                            font = font_manager.FontProperties(family=self.filename, weight='regular') 
                            self.filename = font_manager.findfont(font, fallback_to_default=False, rebuild_if_missing=True)
                        except ValueError as e:
                            try:
                                # rebuild_if_missing doesn't rebuild the cache. This is gross
                                # but actually forces a rebuild.
                                self.filename = font_manager._load_fontmanager(try_read_cache=False).findfont(font, fallback_to_default=False)
                            except:
                                logging.error("Unable to find font %s: %s", filename, e.args)
                                sys.exit(-1)



                self.font = ImageFont.truetype(self.filename, size=self.size)
                for replacement, value in replacements.items():
                    setattr(self, replacement, locals()[replacement] or find_replace(self.filename, value))
                    fname = getattr(self, replacement, None)
                    if fname:
                        setattr(self, replacement.replace("filename", "font"), ImageFont.truetype(fname, size=self.size))

                logging.debug("Using font: %s", self.filename)
                logging.debug("Using bold font: %s", self.boldfilename)
                logging.debug("Using italic font: %s", self.italicfilename)
                logging.debug("Using bolditalic font: %s", self.bolditalicfilename)

            try:
                hyphen_box = self.font.getbbox("‐")
            except:
                hyphen_box = (0,0,0,0)
            if hyphen_box[1] == hyphen_box[3]:
                self.hyphen = "-"
            else:
                self.hyphen = "‐" # Could be "‐" if font supports it.

        else:
            self.font = NonProportional
            self.boldfont = NonProportional
            self.bolditalicfont = NonProportional
            self.italicfont = NonProportional
            self.hyphen="‐"

    def get_space_width(self, space, bold=False, italic=False):
        return self.get_cached_width(space, bold, italic)

    def get_cached_width(self, character, bold=False, italic=False):
        """Return the width of a specified space, caching it for future use.
        Do not call this for non-spaces, or the cache will explode."""
        if character not in self.space_cache:
            self.space_cache[(character, bool(bold), bool(italic))] = self.get_word_width(character, bold, italic)
        return self.space_cache[(character, bool(bold), bool(italic))]

    def get_space_stretch(self, space, bold=False, italic=False):
        """How much is a space allowed to stretch?"""
        return self.get_space_width(space, bold, italic)*3/6

    def get_space_shrink(self, space, bold=False, italic=False):
        """How much is a space allowed to shrink?"""
        return self.get_space_width(space, bold, italic)*3/9

    def get_hyphen_width(self, bold=False, italic=False):
        return self.get_cached_width(self.hyphen, bold, italic)

    def get_word_width(self, string, bold=False, italic=False):
        return self.font.getlength(string)



#!/usr/bin/env python3
import os
from locale import getlocale, getdefaultlocale, setlocale
import locale
import copy
import logging
import argparse

from PIL import Image, ImageDraw
import langcodes

from .font import Font, NonProportional
from .linebreak import BasicWordSplitter, RichWordSplitter, HBox, Glue
from .wordwrap import KnuthPlassWrap, GreedyWrap

"""
   There are 3 different kinds of classes:

   Word-breaker: Breaks a word into HBoxes/Glue/Penalties. This may involve parsing HTML or other rich text
   tags. BasicWordSplitter is plain text. RichWordSplitter is mostly text, but allows <b> and <i>
   tags for bold/italic text.

   Wrapper: Handles laying out broken-down boxes into lines. GreedyWrap does naïve greedy wrapping: cram
   as much as you can onto a line, then break. KnuthPlassWrap implements the Knuth-Plass wrapping algorithm
   used by TeX/LaTeX.

   Font: Handles font loading and metrics.
"""

class TextWrap():
    @classmethod
    def text_wrapper(self, arg_dict):
        if "image" not in arg_dict:
            arg_dict["image"] = "/dev/null"
        arglist = []
        for k, v in arg_dict.items():
            arglist.append(f"--{k}")
            arglist.append(f"{v}")
        print(arglist)
        args  = parse_arguments(args=arglist)
        print(args)
        tw = TextWrap(args)
        return tw

    def __init__(self, args, line_widths=None):

        if args.image:
            self.font = Font(args.font, size=args.font_size)
        else:
            self.font = Font(NonProportional)

        splitter = globals().get(args.splitter+"Splitter")
        logging.info("Using splitter %s", splitter)
        self.splitter = splitter(font=self.font, args=args)

        self.width = args.width
        if self.width is None:
            if not args.image:
                self.width = 80
            else:
                self.width = 600

        line_max = self.width
        self.args = {"line_max": line_max, "font": self.font, "line_widths": line_widths}

        wrapper = globals().get(args.wrapper+"Wrap")
        logging.info("Using wrapper %s", wrapper)
        self.wrapper=wrapper(**self.args)
        self.wrapper.splitter = self.splitter
        self.wrapper.args = self.args

    def render_text(self, text):
        nodes = self.splitter.split(text)
        lines = self.wrapper.wrap_nodes_to_lines(nodes)
        return lines

    def split_and_wrap(self, text):
        nodes = self.splitter.split(text)
        lines = self.wrapper.wrap_nodes(nodes)
        return lines

    def render_image(self, text, filename=None, img=None, where=(0,0), margin=(0,0,0,0), fg=(0,0,0,255), bg=(255,255,255,255)):
        """Render to a PIL Image img, writing with foreground color fg. 

        If img is None, render to a newly allocated correctly sized image 
        with margins of (top, right, bottom,left) and background color bg.

        If img is not None, write text to that image, beginning at offset where.

        If filename is not None, save the img to filename."""

        if (filename is None and img is None) or (filename is not None and img is not None):
            raise ValueError("render_image must specify one of filename or img")
        

        lines = self.split_and_wrap(text)
        font = self.font

        line_offset = round(self.font.size*self.font.linespacing)
        if isinstance(margin, int):
            margin = (margin,)*4

        in_img = img
        if img is None:
            height = line_offset*len(lines) + margin[0]+margin[2]
            width = self.width + margin[1] +margin[3]
            where = (margin[0], margin[1])
            img = Image.new("RGBA", (width,height), bg)

        draw = ImageDraw.Draw(img)
        draw.fontmode="L" #Anti-alias; "1" to turn off

        v_origin = where[0]
        h_origin = where[1]

        for i, line in enumerate(lines):
            h_offset = h_origin
            ratio = line.ratio
            for j, item in enumerate(line.nodes):
                if isinstance(item, HBox):
                    draw.text( (h_offset, i*line_offset + v_origin), item.value, font=item.font, fill=fg)
                    h_delta = item.font.getlength(item.value)
                    h_offset += h_delta
                elif isinstance(item, Glue):
                    h_offset += item.width + (item.stretch* 0 if ratio>0 else item.shrink*ratio)

        if filename is not None:
            img.save(filename)
        if not in_img:
            return img
        return len(lines)*line_offset + v_origin


def parse_arguments(args=None):
    warn_locale = False
    try:
        locale.setlocale(locale.LC_CTYPE, '')
    except locale.Error:
        warn_locale = True
    mylocale = getlocale(locale.LC_CTYPE)[0]
    if mylocale is None:
        mylocale = 'en_US' # Default to US english if we can't figure out the user's preference

    parser = argparse.ArgumentParser("Lay out specified text")
    parser.add_argument("--wrapper", choices=["KnuthPlass", "Greedy"], default="KnuthPlass", help="Which word-wrapping algorithm to use; KnuthPlass is the TeX algoritm, Greedy is simple greedy-matching")
    parser.add_argument("--file", "-f", help="Pull text to wrap from FILE")
    parser.add_argument("--image", "-i", help="Generate image in file named IMAGE. Image filename must end with .png", default=None)
    parser.add_argument("--margin", help="Have a margin of width MARGIN around the text (image mode only)", default=5, type=int)
    parser.add_argument("--hyphenate", "-H", action="store_true", help="Attempt to hyphenate words with PyHyphen.")
    parser.add_argument("--width", "-w", type=int, default=None, help="Width to render. This is in pixels in image mode, or in characters in text mode.")
    parser.add_argument("--font", "-F", help="Font to use. Can be a font face (Calibri, Lucida Grande, Liberation Sans, etc) or a full path to a .ttf or .otf file. If you pass it a regular/normal weight, it'll try to find the bold/italic variants in the same location.", default=None)
    parser.add_argument("--font-size", help="Font size, either points or pixels, e.g. 12pt or 16px", default="12pt")

    less_common = parser.add_argument_group("Advanced Options")
    less_common.add_argument("--splitter", choices=["RichWord", "BasicWord"], default="RichWord", help="Which text splitter to use. RichWord treats <b> and <i> as bold/italic and <nb> for non-breaking runs. BasicWord is a simple plaintext splitter (splits at breaking whitespace only).")
    less_common.add_argument("--no-warn-locale", "-W", action="store_false", dest="warn_locale", help="Suppress warnings about being unable to set locale")
    less_common.add_argument("--verbose", "-v", action="store_true", help="Print extra info")
    less_common.add_argument("--debug", action="store_true", help="Print debugging info")
    less_common.add_argument("--test", "-t", action="store_true", help="Run on test data with text and image output")
    less_common.add_argument("--language", "-l", default=mylocale, help="Language to use for hyphenation (e.g. en_GB, ja_JP, de_DE, fr_FR, zh_HANS, es_ES, etc); defaults to platform locale, you can often set the LC_ALL environment variable or change system configuration to set that.")
    less_common.add_argument("--hyphen-penalty", type=int, default=100, help="Penalty factor for hyphenation (lower results in more hyphenation; default 100)")
    less_common.add_argument("--midbreak-penalty", type=int, default=25, help="Penalty factor for breaking after a hard - or — (lower results in more breaks; default 25)")
    args = parser.parse_args(args)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format="%(asctime)s|%(levelname)s - %(message)s")
    if args.verbose:
        logging.basicConfig(level=logging.INFO, format="%(asctime)s|%(levelname)s - %(message)s")
    else:
        logging.basicConfig(level=logging.WARNING, format="%(asctime)s|%(levelname)s - %(message)s")

    if warn_locale and args.warn_locale and args.hyphenate:
        logging.warning(f"Unable to set locale (bad LC_CTYPE or LANG environment variable?), using {mylocale}. Consider --language if needed. Use --no-warn-locale/-W to suppress this warning.")

    logging.info("Args: %s", str(args))

    if args.hyphenate:
        try:
            import hyphen
        except:
            logging.error("Cannot import hyphen; try 'pip install pyhyphen' if you need hyphenation support")
            sys.exit(-1)

        try:
            hyphen.Hyphenator(args.language)
        except IOError:
            standard = langcodes.standardize_tag(args.language).replace("-", "_")
            try:
                hyphen.Hyphenator(standard)
                args.language = standard
            except IOError:
                logging.error(f"Unable to hyphenate {args.language}")
                sys.exit(-1)

        try:
            human_language = ": " + langcodes.Language.get(args.language).display_name()
        except:
            human_language = ""
        logging.info(f"Using hyphenation language {args.language}{human_language}")

    return args




#!/usr/bin/env python3

import copy

from .pyparagraph import TextWrap, parse_arguments

desc_sample = "But, in a larger sense, we can not dedicate—we can not consecrate—we can not hallow—this ground. The brave men, <b>living and dead</b>, who struggled here, have consecrated it, far above our poor power to add or detract. The world will little note, nor long remember what we say here, but it can never forget what they did here. It is for us the living, rather, to be dedicated here to the unfinished work which they who fought here have thus far so nobly advanced. It is rather for us to be here dedicated to the great task remaining before us—that from these honored dead we take increased devotion to that cause for which they gave the last full measure of devotion—that we here highly resolve that these dead shall not have died in vain—that this nation, under God, shall have a <i>new birth of freedom</i>—and that government of the people, by the people, for the people, shall not perish from the earth."

def main():

    args = parse_arguments()

    if args.test:
        test_args = copy.copy(args)
        test_args.image = False
        tw = TextWrap(args=test_args)
        lines = tw.render_text(desc_sample)
        for line in lines:
            print(line)

    if args.test:
        text_to_wrap = desc_sample
    else:
        text_to_wrap = open(args.file, "r").read()

    if args.image:
        tw = TextWrap(args=args)
        tw.render_image(text_to_wrap, args.image, img=None, margin=args.margin)
    else:
        tw = TextWrap(args=args)
        lines = tw.render_text(text_to_wrap)
        print("\n".join(lines))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from collections import namedtuple
from copy import copy
from itertools import takewhile
import logging
import os

from PIL import ImageFont, Image, ImageDraw

from .linebreak import HBox, Penalty, Infinity, Glue
from .linked_list import LinkedList, Node

LineWidth = namedtuple("LineWidth", ["max", "ideal" ])
class Breakpoint:
    def __init__(self, position, demerits, ratio, line, fitness_class, totals, previous):
        self.position = position
        self.demerits = demerits
        self.ratio = ratio
        self.line = line
        self.fitness_class = fitness_class
        self.totals = totals
        self.previous = previous
    def __repr__(self):
        return f"Breakpoint(position={repr(self.position)}, demerits={repr(self.demerits)}, ratio={repr(self.ratio)}, line={repr(self.line)})"

def new_breakpoint(position, fitness_class, totals, candidate):
    previous = candidate.active if candidate else None
    demerits = candidate.demerits if candidate else 0
    ratio = candidate.ratio if candidate else 0
    totals = totals or Glue(0,0,0)
    line = (previous.data.line + 1) if previous else 0
    return Breakpoint(position, demerits, ratio, line, fitness_class, totals, previous)

Break = namedtuple("Break", ["position", "ratio"])
Candidate = namedtuple("Candidate", ["active", "demerits", "ratio"])
LineData = namedtuple("LineData", ["ratio", "nodes"])
def is_newline(n):
    return isinstance(n, Penalty) and (n.penalty <= -Infinity)

class WrapInfo:
    def __init__(self, font, line_max=80, line_ideal=None, line_widths=None):
        self.line_max = line_max
        if line_ideal is None:
            line_ideal = line_max
        self.line_ideal = line_ideal
        self.line_widths = line_widths
        self.demerits = namedtuple("Demerits", ["line", "flagged", "fitness"])(10,100,300)
        self.tolerance = 2
        self.font = font

    def calc_line_length(self, i):
        i = i - 1 # i is 1-based
        if self.line_widths and len(self.line_widths)>i:
            return LineWidth(max=self.line_widths[i], ideal = self.line_widths[i])
        return LineWidth(max=self.line_max, ideal=self.line_ideal)

class KnuthPlassWrap(WrapInfo):
    def linebreak(self, nodes, lines):
        self.activeNodes = LinkedList()
        self.sum_ = Glue(0,0,0)
        self.nodes = nodes
        breaks = []

        tmp = Node(Candidate(0, Infinity, 0))

        def compute_ratio(node, active, current_line):
            width = self.sum_.width - active.totals.width
            line_length = self.calc_line_length(current_line).ideal

            if isinstance(node, Penalty):
                width += node.width

            def diff_or_infinity(self_val, active_val):
                difference = self_val - active_val
                if difference <= 0:
                    return Infinity
                return (line_length - width) / difference

            if width < line_length:
                return diff_or_infinity(self.sum_.stretch, active.totals.stretch)
            elif width == line_length:
                return 0
            else:
                return diff_or_infinity(self.sum_.shrink, active.totals.shrink)

        def compute_sum(nodes):
            # We're only totalling up until the first HBox or first IP that isn't element 0
            def not_ends_list(node):
                return not (isinstance(node, HBox) or (node is not nodes[0] and is_newline(node)))

            # Add all the Glue nodes in that range to our base result
            return sum((node for node in takewhile(not_ends_list, nodes) if isinstance(node, Glue)), 
                         copy(self.sum_))

        def mainLoop(node, index, nodes, wobble=False):
            active = self.activeNodes[0] if self.activeNodes else None

            while active is not None:
                candidates = [ Candidate(None, Infinity, None)] * 4

                while active:
                    current_line = active.data.line+1
                    ratio = compute_ratio(nodes[index] if index<len(nodes) else None, active.data, current_line)

                    if ratio < -1 or is_newline(node):
                        self.activeNodes.remove(active)

                    if -1 <= ratio and ratio <= self.tolerance:
                        badness = 100 * (abs(ratio)**3)

                        demerits = (self.demerits.line + badness)**2
                        if isinstance(node, Penalty) and node.penalty >= 0:
                            demerits += node.penalty**2
                        elif isinstance(node, Penalty) and node.penalty > -Infinity:
                            demerits -= node.penalty**2

                        # Try to avoid having 2 flagged lines in a row (flagging usually means
                        # hyphenated)
                        if isinstance(node, Penalty) and len(nodes)>active.data.position and isinstance(nodes[active.data.position], Penalty):
                            demerits += self.demerits.flagged * node.flagged * nodes[active.data.position].flagged

                        class_cutoffs = { 0:-.5, 1:.5, 2:1, 3:Infinity+ratio}
                        current_class = min( k for k in class_cutoffs if ratio < class_cutoffs[k])

                        if abs(current_class - active.data.fitness_class) > 1:
                            demerits += self.demerits.fitness

                        demerits += active.data.demerits
                        
                        if demerits < candidates[current_class].demerits:
                            candidates[current_class] = Candidate(active, demerits, ratio)

                    active = active.next

                    if active is not None and active.data.line >= current_line:
                        break

                current_sum = compute_sum(self.nodes[index:])

                for fitness_class, candidate in enumerate(candidates):
                    if candidate.demerits < Infinity:
                        new_node = Node(new_breakpoint(index, fitness_class, current_sum, candidate))
                        if active is not None:
                            self.activeNodes.insertBefore(active, new_node)
                        else:
                            self.activeNodes.append(new_node)

        self.activeNodes.append(Node(new_breakpoint(0,0,None,None)))

        last = None

        for i, node in enumerate(self.nodes):
            if isinstance(node, HBox):
                self.sum_.width += node.width
            elif isinstance(node, Glue):
                if isinstance(last, HBox):
                    mainLoop(node, i, self.nodes, wobble=True)
                self.sum_.width += node.width
                self.sum_.stretch += node.stretch
                self.sum_.shrink += node.shrink
            elif isinstance(node, Penalty) and node.penalty < Infinity:
                mainLoop(node, i, self.nodes)
            last = node
        
        if self.activeNodes:
            tmp = min(self.activeNodes)

            while tmp is not None:
                breaks.append( Break(tmp.data.position, tmp.data.ratio))
                tmp = tmp.data.previous

            breaks.reverse()
            return breaks
        return []

    def wrap_nodes(self, nodes):
        for tolerance in (1,2,3):
            logging.debug("Attempting to find fit with tolerance %d", tolerance)
            self.tolerance=tolerance
            nodes = copy(nodes)
            my_breaks = self.linebreak(nodes, None)
            if my_breaks:
                break
        else:
            logging.error("Cannot find a Knuth-Plass wrap, falling back to Greedy wrapping")
            wrapper = GreedyWrap(**self.args)
            return wrapper.wrap_nodes(nodes)

        lines = []
        line_start = 0
        for b in my_breaks[1:]:
            point = b.position
            r = b.ratio
            for j, n in enumerate(nodes[line_start:]):
                if isinstance(n, HBox) or is_newline(n):
                    line_start += j
                    break
            need_hyphen=False
            if isinstance(nodes[point], Penalty) and nodes[point].flagged and nodes[point].penalty > -Infinity:
                need_hyphen = True
                for n in nodes[point-1:line_start:-1]:
                    if isinstance(n, HBox):
                        need_hyphen = n.value[-1].isalpha()
                        which_font = n.font
                        break
            need_hyphen=[HBox(self.font.hyphen, which_font)] if need_hyphen else []
            lines.append(LineData(r, nodes[line_start:point+1]+need_hyphen))
            line_start = point

        return lines
    def wrap_nodes_to_lines(self, nodes):
        lines = self.wrap_nodes(nodes)

        text_lines = []
        for line in lines:
            text_lines.append([])
            indent = False
            spaces=0
            totalAdjustment = 0
            wordSpace = 10 #FIXME
            for node in line.nodes:
                if isinstance(node, HBox):
                    text_lines[-1].append(node.value)
                elif isinstance(node, Glue):
                    text_lines[-1].append(" "*(1 if node.width else 0))
        text_lines = [ "".join(t) for t in text_lines ]
        return text_lines

class GreedyWrap(WrapInfo):
    def wrap_nodes(self, nodes):
        lines = [[]]
        am_hyphenating = False
        for node in nodes:
            # No penalties in greedy wrapping
            if not isinstance(node, (Glue, HBox)):
                if node.width:
                    am_hyphenating = True
                continue
            if not lines[-1]:
                lines[-1].append(node)
            else:
                line_len = sum(node.width for node in lines[-1])
                if line_len + node.width < self.calc_line_length(len(lines)-1).ideal:
                    lines[-1].append(node)
                else:
                    if am_hyphenating:
                        if lines[-1] and isinstance(lines[-1][-1], HBox) and lines[-1][-1].value[-1].isalpha():
                            lines[-1].append(HBox(self.font.hyphen, self.font.font))
                    if isinstance(node, Glue):
                        lines.append([])
                    else:
                        lines.append([node])
                am_hyphenating = False
        return [LineData(1, line) for line in lines]

    def wrap_nodes_to_lines(self, nodes):
        lines = self.wrap_nodes(nodes)
        for i, line in enumerate(lines):
            lines[i] = "".join( v.value if isinstance(v, HBox) else " "*(1 if v.width else 0) for v in line.nodes)
        return lines
